/* Correlate routines for uushred 4.2pl3  by Mark Neidengard
   (mneideng@ugcs.caltech.edu)  Circa 6-1996 */


/* this function checks to see if <string> represents a subject line from a 
   multipart uue...it looks for ##/## or ## of ## somewhere in the line...
   it returns the index of the first character that is part of these
   constructs */

int qualify (string)
char string[];
{
  int position=0, len=strlen (string);
  while (position<len) {
    if (string[position] >= '0' && string[position] <= '9') {
      position+=2;
      if (string[position] >= '0' && string[position] <= '9') ++position;
      if (position<len && string[position]=='/') {
        ++position;
        if (position<len && string[position] >= '0' && string[position] <= '9')
          return position-3;
      }
      if (position<len && !strncmp(string+position," of ",4)) {
        position +=4;
        if (position<len && string[position] >= '0' && string[position] <= '9')
          return position-6;
      }
      continue;
    }
    ++position;
  }
  return -1;
}

/* this function takes <start> as the position in the string where the 
   part/whole or part of whole stuff is.  It takes this and extracts the
   total number of expected parts, and returns that */
int find_total (string,start)
char string[];
int start;
{
   int result;
   start+=2;
   if (string[start] >= '0' && string[start] <= '9') ++start;
   if (string[start]=='/') ++start;
   else start+=4;
   sscanf (string+start,"%d",&result);
   return result;
}

/* this function takes string and tries to see if the part it represents is
   next in order to <part_no> */
int sequitur (string,part_no,first)
char string[];
int part_no,first;
{
/*   int number=(10*(string[first]-48)+(string[first+1]-48)); */
   int number;
   sscanf (string+first,"%d",&number);
   switch (number-part_no) {
      case 0: return -1;break;
      case 1: return 1; break;
      default: return 0; break;
   }
}

/* take the array of subjects and mark from start to offset as uncorrelated */
backdate (the_array,start,offset)
struct subject_line the_array[];
int start, offset;
{
   while (offset >= 0) the_array[start+offset--].correlated=1;
}


/* the main corellation engine.  Takes the array of subject structs and a
   highest subject number */
correlate (sub_array,last)
struct subject_line sub_array[];
int last;
{
   int sub_counter=0;
   int first;
   int total;
   int part_no;
   int offset=0;
   int result; /* for sequitur() */

   while (sub_counter<last) {
      offset=0; part_no=0;
      if (sub_array[sub_counter].correlated == 1 ||
          (first=qualify (sub_array[sub_counter].string))==-1) {
         ++sub_counter;
         continue;
      }
      /* if we don't have the first part, or if we have part "0", mark it
         invalid and continue */
      if (strncmp (((sub_array[sub_counter].string)+first),"01",2)) {
         sub_array[sub_counter++].correlated=1;
         continue;
      }
      total=find_total (sub_array[sub_counter].string,first);
      part_no=1;
      while ((sub_counter+part_no+offset)<last && qualify(sub_array[sub_counter+part_no+offset].string)==first && sub_array[sub_counter+part_no+offset].correlated != 1) {
         if (strncmp (sub_array[sub_counter].string,sub_array[sub_counter+part_no+offset].string,first))
            break;
         /* that is, we have the first element of a multipart uue; we now check
            to see if the next subject matches as far as the part number; if
            not, break */
         if (!(result=sequitur (sub_array[sub_counter+part_no+offset].string,part_no,first)))
            break;
         if (result==-1) 
            sub_array[sub_counter+part_no+offset++].correlated=1;
         else ++part_no;
         /* if result == -1, then the subject in question was a duplicate of
            the previous one, so mark it as invalid; else, we have the next
            subject in order, so increment the part_no */
      }
      if (part_no > 0 && part_no != total) backdate (sub_array,sub_counter,part_no+offset-1);

      sub_counter+=(part_no+offset);
   }
}
