/* Unix-Unix Simple Heirarchial Reordering Decoder version 4.2pl3
   Written by Mark Neidengard, mneideng@ugcs.caltech.edu
   uuconvert routines by permission of weigley@sun.acsu.buffalo.edu
   June, 1996.  Permission to distribute this code freely as long as
   source remains unchanged.  Feel free to make any modifications 
   you want to your personal copy, and if you feel you have a patch or
   feature others would profit by, don't heistate to let me know.
   Please submit all suggestions to me via email.   */

#define TRUE    1
#define FALSE   0
#define DEC(c)  (((c) - ' ') & 077)

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>     /* added for compatibility w/ DOS and OS/2 systems */
FILE *out;
#include "source/structs.h"
#include "source/initargs.c"
#include "source/handleargs.c"
#include "source/messages.c"
#include "source/kill.c"
#include "source/subject.c" 
#include "source/correlate.c"
#include "source/decode.c"

 
int main (argc, argv)   /* note, "old style" syntax for compatibility */
int argc;
char *argv[];
{
 
   /*  all internal variable definitions go here */

struct switches main_switch; /* struct holds status of program options */
int tmp;
int arg_count;               /* what argument number we're on */
int subject_no=0;            /* the last subject in the list */
int read_me=1;               /* toggles whether we store subjects */
int out_count=0;             /* for output subject counting */
int counter=1;               /* for keeping track of ends of articles */
int tmp_file_pos;            /* for noting our places in the input files */
char seeking,tmp2;           /* used for part 1-begin / part N-end detect */
FILE *infile=NULL;           /* file pointers */ 
FILE *outfile=NULL;          /* file pointers */ 
int kill_number;             /* the number of lines in the killfile */
char pipe_name[MAXBUFSIZE];  /* the name of the temp. file for pipeline-in */
char **kill_array;           /* for kill() */
char buffer[MAXBUFSIZE];     /* for fgets() et. al. */
struct subject_line *subject_index;  /* pointer to the start of the subj list*/

/* make sure subject list is blank */
subject_index=(struct subject_line *)calloc(MAXSUB,sizeof(struct subject_line));


  initargs(&main_switch); /* set all switches off */
 
  if (argc < 2 ) {  /* they ran without any arguments */
    help_screen();
    exit (1);
  }

  if ((arg_count=handleargs(argc,argv,&main_switch))==argc)
    if (!main_switch.pipeline) {
      no_files(argv[0]);

      /* that is, if all the arguments are parameters and not input file
         specifications, and if we're not in pipeline read mode (--),
         then tell the user no files were specified. */

      exit (2);
    }

  /* if we are using a killfile, let's build the array of subjects to kill */
  if (main_switch.killfile) build_kill (&kill_array,&main_switch,&kill_number);

  /* now the fun begins...we start reading data from the inputs */
  infile=0;      /* make sure we don't have a file open yet */
  seeking=0;     /* we aren't looking for "begin" or "end" yet */

  if (main_switch.phase_info) {
    printf ("Getting subjects...\n");
  }

  do {
    if (infile) fclose(infile);
    if (arg_count==argc) {        /* we're in pipeline mode */
      infile=stdin;

      /* now we use some trickery to get the tmp file to work.  we will 
         spoof uushred into thinking that the tmp file is one of the
         uue source files, so that when the subject entries store which
         input file they came from, they will reference the tmp file.
         we do this by trashing argv[argc-1], which is valid for
         pipeline mode because the user had to specify at least one
         commandline switch (ie arg >= 2) */

      --arg_count;
      sprintf(pipe_name,"%s%06d.tmp",TMP_PREFIX,rand());
      argv[arg_count]=pipe_name;

      if (outfile=fopen(pipe_name,"r")) {     /* check to see if the tmp */
        file_error (argv[0],pipe_name,"pc");  /* file already exists */
        exit(6);
      }
      fclose(outfile);
      if ((outfile=fopen(pipe_name,"wb"))==NULL) {
        file_error (argv[0],pipe_name,"pc");
        exit(6);
      }
    }
    else 
      if ((infile=fopen (argv[arg_count],"r"))==NULL) {
        file_error(argv[0],argv[arg_count],"r");
        exit (3);
      }

    for(;;) {
      /* initialize tmp_file_pos */
      if (main_switch.pipeline) tmp_file_pos=ftell(outfile);
      else tmp_file_pos=ftell(infile);

      /* trap on end of file conditions, after terminating the last subject
         entry */
      if (!fgets (buffer,MAXBUFSIZE-1,infile)) {
        if (subject_no > 0) subject_index[subject_no-1].finish=counter;
        break;
      }
      ++counter;

      if (main_switch.pipeline) fputs(buffer,outfile);

      /* if we're seeking, examine the buffer...*/
      if (main_switch.correlate && seeking) {
        if ((seeking == 'a' || seeking == 'b') && !strncmp(buffer,"begin ",6)) {
          if (seeking == 'a') seeking='e';
          else seeking=0;
        }
        else if (seeking == 'e' && !strncmp(buffer,"end",3)) seeking=0;
      }

      if (!strncmp (buffer,"Subject:",7)) {

        /* terminate the previous subject entry if we haven't already */
        if (subject_no > 0 && !subject_index[subject_no-1].finish) 
          subject_index[subject_no-1].finish=counter-1;

        /* groom the subject and compare it against "kill" phrases.  Return
           1 if the article is still ok to read */

        read_me=subject_handle (buffer,subject_index,subject_no, &main_switch,kill_number,kill_array);

        if (read_me) {       
          subject_index[subject_no].start=tmp_file_pos;
          subject_index[subject_no].argnum=arg_count;

          counter=1;

          /* if seeking != 0, then the previous subject was wack */
          if (seeking) subject_index[subject_no-1].correlated=1,seeking=0;

          /* if we're part 1, start looking for "begin"; if we're part N,
             start looking for "end" */
          if (main_switch.correlate) {
            tmp=qualify(subject_index[subject_no].string);
            tmp2=atoi(subject_index[subject_no].string + tmp);
            seeking=0;
            if (tmp2 == 1) seeking='b';
            if (tmp>=0 && tmp2==find_total(subject_index[subject_no].string,tmp)) {
              if (tmp2==1) seeking='a';
              else seeking='e';
            }
          }
          if (main_switch.input_sub) printf ("%s",buffer);
        }
        subject_no+=read_me;     /* increment the number of subjects we
                                    have if subject_handle() gave us 1;
                                    outside of if {} to expose parallelism */ 
      }
    }
  }
  while (++arg_count < argc);

  if (!subject_no) {
    printf ("%s: No subjects present\n",argv[0]);
    exit (4);
  }

  /* let's make sure that the tmp file is closed, if we used one */
  if (main_switch.pipeline) fclose (outfile);

  /* having read the data, we sort the subjects with quicksort */
  if (main_switch.phase_info) {
    printf ("Sorting...\n");
  }

  quicksort (subject_index,0,subject_no-1);

  /* and then, we correlate them.. */
  if (main_switch.correlate) {
    if (main_switch.phase_info) {
      printf ("Correlating...\n");
    }
      correlate (subject_index,subject_no);
  }

  /* show(?) the sorted subject list */
  if (main_switch.sorted_sub) {
    while (out_count < subject_no) {
      if (subject_index[out_count].correlated) printf ("---> ");
      printf ("%s",subject_index[out_count].string);
      ++out_count;
    }
  }

  /* and finally, we output using the subject_index as a guide */

  out_count=0;
  if (main_switch.phase_info) {
    printf ("Output phase...\n");
  }

  if (main_switch.correlate) {

    if (!(*(main_switch.uncorrelated_out)))
      outfile=stdout;
    else {
      if ((outfile=fopen (main_switch.uncorrelated_out,"ab"))==NULL)
        fprintf(stderr,"%s: Cannot send uncorrelated output to %s\n", argv[0],main_switch.uncorrelated_out);
      if (main_switch.phase_info) {
         printf ("Unmatched parts to ");
         if (!(*(main_switch.uncorrelated_out)))
            printf ("stdout\n");
         else printf ("%s\n",main_switch.uncorrelated_out);
      }
    }

    if ((infile=fopen(argv[subject_index[out_count].argnum],"r"))==NULL) {
      file_error(argv[0],argv[subject_index[out_count].argnum],"r");
      exit (3);
    }

    fseek(infile,subject_index[0].start,0);

    for (;;) {
      counter=0;
      if (subject_index[out_count].correlated)
        while (counter != subject_index[out_count].finish) {
          fgets(buffer,MAXBUFSIZE-1,infile);
          fprintf (outfile,"%s",buffer);
          ++counter;
        }

      ++out_count;

      if (out_count >= subject_no) break;

      if (subject_index[out_count].argnum != subject_index[out_count-1].argnum) {
        fclose(infile);
        if ((infile=fopen(argv[subject_index[out_count].argnum],"r"))==NULL) {
          file_error(argv[0],argv[subject_index[out_count].argnum],"r");
          exit (3);
        }
      }
      fseek(infile,subject_index[out_count].start,0);
    }
  }
  if (infile && infile != stdin) fclose(infile);
  if (outfile && outfile != stdout) fclose (outfile);

  if (main_switch.no_decode) {
    out_count=0;
    if (!(*(main_switch.undecoded_out)))
      outfile=stdout;
    else if ((outfile=fopen (main_switch.undecoded_out,"ab"))==NULL) {
        file_error(argv[0],main_switch.undecoded_out,"w");
        exit(5);
    }
    if (main_switch.phase_info) {
      printf ("Undecoded uue's to ");
      if (!(*(main_switch.uncorrelated_out)))
        printf ("stdout\n");
      else printf ("%s\n",main_switch.undecoded_out);
    }

    if ((infile=fopen(argv[subject_index[out_count].argnum],"r"))==NULL) {
      file_error(argv[0],argv[subject_index[out_count].argnum],"r");
      exit (3);
    }
    fseek(infile,subject_index[0].start,0);

    for (;;) {
      counter=0;
      if (!subject_index[out_count].correlated) 
        while (counter != subject_index[out_count].finish) {
          fgets(buffer,MAXBUFSIZE-1,infile);
          fprintf(outfile,"%s",buffer);
          ++counter;
        }

      ++out_count;
 
      if (out_count >= subject_no) break;
 
      if (subject_index[out_count].argnum != subject_index[out_count-1].argnum) {
        fclose(infile);
        if ((infile=fopen(argv[subject_index[out_count].argnum],"r"))==NULL) {
          file_error(argv[0],argv[subject_index[out_count].argnum],"r");
          exit (3);
        }
      }
      fseek(infile,subject_index[out_count].start,0);
    }
    if (infile && infile != stdin) fclose(infile);
    if (outfile && outfile != stdout) fclose (outfile);
  }

  /* here is the output/decode section, selected if ! no_decode */
  else {
    char *s2, *s1, *s0, *tmp_s;
    int length;
    static char s[3 * (MAXBUFSIZE)];
    static int echo_on = FALSE, started = FALSE;
    static int line_length = 0, lines_to_go = 0, firstline=FALSE;
 
    s0 = s;
    s1 = s0 + (MAXBUFSIZE);
    s2 = s1 + (MAXBUFSIZE);
 
    out_count=0;
    s0[0] = s1[0] = s2[0] = '\0';  /* Clear strings */

    if ((infile=fopen(argv[subject_index[out_count].argnum],"r"))==NULL) {
      file_error(argv[0],argv[subject_index[out_count].argnum],"r");
      exit(3);
    }
    fseek(infile,subject_index[0].start,0);
 
    for (;;) {
      counter=0;
      if (!subject_index[out_count].correlated)
        while (counter != subject_index[out_count].finish) {
          fgets(buffer,MAXBUFSIZE-1,infile);

          /* patch ^M handling for non DOS machines */
#ifndef DOS
          length=strlen(buffer)-1;
          if (buffer[length-1] == 13) {    /* ^M is 013 decimal */
            buffer[length-1] = buffer[length];
            buffer[length] = '\0';
          }
#endif

          strcpy (s0,buffer);
          ++counter;

          s0[MAXBUFSIZE-1] = '\0';  /* Make sure string is terminated */
 
          if (strncmp(s0, "begin ", 6) == 0) started=0;
 
          if (!started) {
            if (strncmp(s0, "begin ", 6) == 0) {
              int mode;
              char dest[128];
              firstline=TRUE;
              started = echo_on = TRUE, line_length = 0, lines_to_go = 0;
              sscanf(s0, "begin %o %s", &mode, dest);

              /* turn off all pathname handling.  disallow filenames
                 containing either '/' or '\' characters */
              
              for(length=strlen(dest)-1;length>=0;length--) {
                if (dest[length] == '\\' || dest[length] == '/') break;
              }
              if (length >= 0) strcpy(dest,dest+length+1);
              

              if (out) fclose(out);
              out=fopen(dest, "r");  /* check to see if target file exists */
              if (out != NULL) {
                strcat (dest,".new");
                fclose (out);
              }
              if ((out=fopen(dest, "w"))==NULL) {
                file_error(argv[0],dest,"w");
                exit(5);
              }
              if (main_switch.files_made) printf ("%s\n",dest);
            }
          }
          else {  /* started */
            length = strlen(s0);
            if (line_length == 0) line_length = length;
 
            if (echo_on) {
              lines_to_go = 0;
              if (s0[0] != 'M' || length != line_length) {
                echo_on = FALSE;
                lines_to_go = 2;  /* Lines to go before 'end' is expected */
                if (s0[0] == ' ' || s0[0] == '`') lines_to_go = 1;
              }
            }
            else if (s0[0] == 'M' && length == line_length) echo_on = TRUE;
            else if (lines_to_go > 0)
              if (lines_to_go == 2)
                if (s0[0] == ' ' || s0[0] == '`')
                  lines_to_go = 1;
                else
                  lines_to_go = 0;  /* Unexpected line, so break off */
              else if (lines_to_go == 1)
                if (!strcmp(s0, "end\n")) {
                  doaline(s2);
                  doaline(s1);
                  lines_to_go = 0;  /* Done. Break off */
                  started = FALSE;
                }
                else lines_to_go = 0;  /* Unexpected line, so break off */
          }
    
          if (echo_on && !firstline) doaline(s0), s0[0] = '\0';
          else if (firstline) firstline=FALSE;
    
          tmp_s = s2, s2 = s1, s1 = s0, s0 = tmp_s;
        }
      ++out_count;
      if (out_count>=subject_no) break;

      if (subject_index[out_count].argnum != subject_index[out_count-1].argnum) {
        fclose(infile);
        if ((infile=fopen(argv[subject_index[out_count].argnum],"r"))==NULL) {
          file_error(argv[0],argv[subject_index[out_count].argnum],"r");
          exit(3);
        }        
      }
      fseek(infile,subject_index[out_count].start,0);
    }
  }
  /* here we are, at the end of the oddessy through uue land. now it ends */
  if (main_switch.pipeline)  remove(pipe_name);
  exit (0);
  return 0;
}
